/*
 * Copyright (c) 2000-2023 nVoq Incorporated
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */


import Foundation

//protocol definition so our view controller knows how to communicate with us
protocol AudioRecorderStateDelegate: AnyObject {
    func onAudioRecordingStart()
    func onAudioRecordingStop()
    func onAudioRecordingData(audioData: NSData)
    func onAudioQueueDisposed()
    func onAudioRecordingFailed()
}

protocol DictationHelperDelegate: AnyObject {
    func onTextReceived(text: String)
}
class DictationHelper: AudioRecorderStateDelegate{

    weak var delegate: DictationHelperDelegate?
    private var dictationId = ""
    private var webSocketTask: URLSessionWebSocketTask?
    var connected = false
    var dictationStarted = false
    var oggEncoder: OggEncoder?

    // audio recording has been started.  Initialize the
    // ogg encoder, connect to the dictation service and send a
    // start dictation request
    func onAudioRecordingStart() {
        let quality = Float(1)/10
        oggEncoder = OggEncoder.init(format: 1, rate: Int32(16000), quality: quality)
        if !connected{
            connect()
            sendStartDictation()
        }
    }
    
    // Receive the audio data and send it to the server
    func onAudioRecordingData(audioData: NSData) {
        if (!dictationStarted || !connected){
            return
        }
        guard let oggData = oggEncoder?.encode(audioData as Data) else { return }
      let audioDataCopy = NSData.init(data: oggData)
      sendData(someData: audioDataCopy)
    }

    func onAudioQueueDisposed() {
    }
    
    func onAudioRecordingFailed() {
    }
    
    //make a WebSocket connection to the dictation service
    //and specify a medical specialty - in this case General Medicine
    func connect() {
        let url = URL(string: "wss://test.nvoq.com/wsapi/v2/dictation/topics/general_medicine")!
        webSocketTask = URLSession.shared.webSocketTask(with: url)
        webSocketTask?.receive(completionHandler: onReceive)
        webSocketTask?.resume()
        connected = true
    }
    
    // Send a String (usually a service method call)
    // to the dictation service
    func sendString(text: String) {
        let message = URLSessionWebSocketTask.Message.string(text)
        webSocketTask?.send(message) { error in
            if let error = error {
                print("Error sending message", error)
            }
        }
    }
    
    // Send a binary data to the dictation service
    // usually audio
    func sendData(someData: NSData) {
        let message = URLSessionWebSocketTask.Message.data(someData as Data)
        webSocketTask?.send(message) { error in
            if let error = error {
                print("Error sending message", error)
            }
        }
    }

    // audio recording has stopped.  send a message to the
    // dictation service that we are done sending audio
    func onAudioRecordingStop() {
        print ("audio recording stopped.")
        sendAudioDone()
    }
    
    // Disconnect the WebSocket connection to
    // the dictation service
    func disconnect() {
        print ("disconnecting")
        connected = false
        webSocketTask?.cancel(with: .normalClosure, reason: nil)
    }

    // Verify the incoming message and hand it off for processing
    func onReceive(incoming: Result<URLSessionWebSocketTask.Message, Error>) {
        print ("on receive!")

        if case .success(let message) = incoming {
            print("received \(message) ")
            onMessage(message: message)
        }
        else if case .failure(let error) = incoming {
            print("Error", error)
        }
        if (connected){
            webSocketTask?.receive(completionHandler: onReceive)
        }

    }
    
    // Process the incoming messages.  For this example, we only
    // watch for messages indicating a dictation has started and
    // the text done (dictation finished) messages.
    func onMessage(message: URLSessionWebSocketTask.Message) {
        
        switch message {
            case .string(let text):
                print("Received text message: \(text)")
                if text.contains("STARTDICTATION"){
                    dictationStarted = true
                }
                if (text.contains("\"textDone\":true")){
                    disconnect()
                }
            
                let info = convertStringToDictionary(text: text)
                let data = info?["data"] as? [String:Any]
                if let dictation = info?["id"] as? String {
                  dictationId = dictation
                }
                
                if let kind = data?["kind"] as? String,
                   kind == "HYPOTHESISTEXT",
                    let dictationText = data?["text"] as? String,
                    !dictationText.isEmpty {
                  DispatchQueue.main.async {
                    self.delegate?.onTextReceived(text: dictationText)
                  }
                }
            case .data(let data):
                   print("Received binary message: \(data)")
            @unknown default:
                   fatalError()
        }
    }
    deinit {
        connected = false
        disconnect()
    }
    
    
    /// TODO: Set your user in id and password in authorization field
    func sendStartDictation(){
        //Start dictation operation
        let startDictation: [String: Any] = [
            "apiVersion": "1.1",
            "method": "STARTDICTATION",
            "params": [
                "id": "YOUR_USERNAME",
                "authorization": "YOUR_PASSWORD",
                "externalId": "MyDictation",
                "audioFormat":[
                    "encoding":"ogg",
                    "sampleRate": 16000
                ],
                "snsContext":[
                    "dictationContext":"",
                    "selectionOffset":0,
                    "selectionEndIndex":0
                ],
                "returnSubscriptions": [
                    "HYPOTHESISTEXT",
                    "STABLETEXT"
                ]
            ]
        ]
        do {
            let message =  try JSONSerialization.data(withJSONObject: startDictation, options: JSONSerialization.WritingOptions.prettyPrinted)
            guard let messageString = String(data: message, encoding: String.Encoding.utf8) else {
                print ("null message string")
                return
            }
            sendString(text: messageString)
        }catch  let myError {
            print(myError)
        }

    }
    
    func sendAudioDone(){
        let audioDoneMessage: [String: Any] = [
            "apiVersion":"1.1",
            "method":"AUDIODONE"
        ]
        
        do {
            let message =  try JSONSerialization.data(withJSONObject: audioDoneMessage, options: JSONSerialization.WritingOptions.prettyPrinted)
            guard let messageString = String(data: message, encoding: String.Encoding.utf8) else {
                print ("null message string")
                return
            }
            sendString(text: messageString)
        }catch  let myError {
            print(myError)
        }
    }
    
    func convertStringToDictionary(text: String) -> [String:AnyObject]? {
        if let data = text.data(using: .utf8) {
            do {
                let json = try JSONSerialization.jsonObject(with: data, options: .mutableContainers) as? [String:AnyObject]
                return json
            } catch {
                NSLog("Something went wrong")
            }
        }
        return nil
    }
}
