//
//  NVOQOggEncoder.m
//  SayItSDK
//
/*
 * Copyright (c) 2000-2023 nVoq Incorporated
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */


#import "NVOQOggEncoder.h"

#include <vorbis/vorbisenc.h>
#include <memory>
#include <time.h>

@interface OggEncoder()

// Private struct
typedef struct OggVorbisEncoderContext {
	ogg_stream_state os; /* take physical pages, weld into a logical stream of packets */
	ogg_page         og; /* one Ogg bitstream page.  Vorbis packets are inside */
	ogg_packet       op; /* one raw packet of data for decode */
	vorbis_info      vi; /* struct that stores all the static vorbis bitstream settings */
	vorbis_comment   vc; /* struct that stores all the user comments */
	vorbis_dsp_state vd; /* central working state for the packet->PCM decoder */
	vorbis_block     vb; /* local working space for packet->PCM decode */
    
	int              isHeaderWritten; /* header written yet? */
    
	Byte * accumulator;
	long dataSize;
	long bufferSize;
} OggVorbisEncoderContext;

-(NSString *)getErrorMessage;
-(void)appendAccumulator:(OggVorbisEncoderContext *)ctx data:(char *)data size:(long)size;
-(int)getAccumulatedBytes:(int64_t)encoderContext outBuffer:(char *)outBuffer maxSize:(int)maxSize;
-(int64_t)newEncoderContext:(int)channels rate:(int)rate quality:(float)quality;
-(void)encodePcm:(int64_t)encoderContext pcmByteArray:(char *)pcmByteArray srcOffset:(int)srcOffset srcLength:(int)srcLength;
-(void)deleteEncoderContext:(int64_t)jEncoderContext;

@end

static const int BUF_SIZE = 65535;

@implementation OggEncoder

@synthesize ptr;

- (id)init
{
    return [self initWithFormat:1 rate:16000 quality:0.5f];
}

- (id)initWithFormat:(int)channels rate:(int)rate quality:(float)quality
{
    if(self = [super init]) {
        ptr = [self newEncoderContext:channels rate:rate quality:quality];
    }
	return self;
}

- (NSData *)encode:(NSData *)audioData {
    char * buf = (char *)malloc(BUF_SIZE);
    @try {
        if (audioData != nil) {
            [self encodePcm:ptr pcmByteArray:(char *)[audioData bytes] srcOffset:0 srcLength:(int)audioData.length];
        }
        else {
            [self encodePcm:ptr pcmByteArray:(char *)[audioData bytes] srcOffset:0 srcLength:-1];
        }
        int count = [self getAccumulatedBytes:ptr outBuffer:buf maxSize:BUF_SIZE];
        return [[NSData alloc]initWithBytes:buf length:count];
    }
    @catch (NSException *e) {
        NSString *msg = [NSString stringWithFormat:@"%@ %@", @"Ogg encoding error", [self getErrorMessage]];
        [self throwError:msg];
    }
    @finally {
        if (buf != NULL) {
            free(buf);
        }
    }
}

- (void)dealloc {
    if (ptr != 0) {
        try {
            [self deleteEncoderContext:ptr];
        } catch (NSException *e) {}
        ptr = 0;
    }
}

NSString * lastError = @"";

// Get the last thrown error message
- (NSString *)getErrorMessage {
	return lastError;
}

// Throwing C++ exception here will also cause an exception on the C# side
// But the C# exception will come without a meaningful message
// That is why we set the message here in a variable which can be then read from the C# code in the exception handler
// This is not thread safe, but it shouldn't happen often
// And since the Ogg encoding is used on a data being recorded from a sound card, only one recording can take place at the same time
- (void) throwError:(NSString *)msg {
	lastError = msg;
	@throw [NSException exceptionWithName:@"OggVorbisException" reason:msg userInfo:nil];
}



// Private function
// Accumulate ogg generated data
- (void)appendAccumulator:(OggVorbisEncoderContext *)ctx data:(char *)data size:(long)size {
	if (ctx->bufferSize > 100000000)
		[self throwError:@"Buffer cap reached"];
	if (ctx->dataSize + size > ctx->bufferSize) {
		// Expand the accumulator buffer
		Byte *b = new Byte[ctx->bufferSize + size];
		memcpy(b, ctx->accumulator, ctx->bufferSize);
		delete [] ctx->accumulator;
		ctx->accumulator = b;
		ctx->bufferSize += size;
	}
    
	// Append new data
	memcpy(ctx->accumulator + ctx->dataSize, data, size);
	ctx->dataSize += size;
}

// Read accumulated data
// Function returns number of bytes
-(int)getAccumulatedBytes:(int64_t)encoderContext outBuffer:(char *)outBuffer maxSize:(int)maxSize {
	lastError = @"";
    
	OggVorbisEncoderContext *ctx = (OggVorbisEncoderContext *)encoderContext;
	if (!ctx)
		[self throwError:@"Invalid context"];
	int size = maxSize;
	if (size > ctx->dataSize)
		size = (int)ctx->dataSize;
	if (size <= 0)
		return size;
	memcpy(outBuffer, ctx->accumulator, size);
	if (size == ctx->dataSize) {
		ctx->dataSize = 0;
	} else {
		// Shift the data, this might never happen if the outBuffer is large enough
		ctx->dataSize -= size;
		memmove(ctx->accumulator, ctx->accumulator + size, ctx->dataSize);
	}
	return size;
}

// Create a new context
-(int64_t)newEncoderContext:(int)channels rate:(int)rate quality:(float)quality {
	lastError = @"";
    
	OggVorbisEncoderContext *ctx = new OggVorbisEncoderContext();
	if (!ctx) {
		[self throwError:@"Allocation error"];
		return 0;
	}
    
	vorbis_info_init(&ctx->vi);
    
	int ret = vorbis_encode_init_vbr(&ctx->vi, channels, rate, quality);
	if (ret)
	{
		[self throwError:@"vorbis_encode_init_vbr error"];
		return 0;
	}
    
	vorbis_comment_init(&ctx->vc);
	vorbis_comment_add_tag(&ctx->vc, "ENCODER", "nVoq");
	vorbis_analysis_init(&ctx->vd, &ctx->vi);
	vorbis_block_init(&ctx->vd, &ctx->vb);
	srand((unsigned int)time(NULL));
	ogg_stream_init(&ctx->os, rand());
    
	ctx->bufferSize = BUF_SIZE;
	ctx->accumulator = new Byte[ctx->bufferSize];
	ctx->dataSize = 0;
    
	return (int64_t) ctx;
}

// Encode PCM data to OGG
-(void)encodePcm:(int64_t)encoderContext pcmByteArray:(char *)pcmByteArray srcOffset:(int)srcOffset srcLength:(int)srcLength {
    
	lastError = @"";
    
	OggVorbisEncoderContext *ctx = (OggVorbisEncoderContext *)encoderContext;
	if (!ctx)
		[self throwError:@"Invalid context"];
    
	int ret = 0;
    
	if (!ctx->isHeaderWritten) {
		ogg_packet header;
		ogg_packet header_comm;
		ogg_packet header_code;
        
		vorbis_analysis_headerout(&ctx->vd, &ctx->vc, &header, &header_comm, &header_code);
        
		ogg_stream_packetin(&ctx->os, &header);
		ogg_stream_packetin(&ctx->os, &header_comm);
		ogg_stream_packetin(&ctx->os, &header_code);
        
		while (true) {
			ret = ogg_stream_flush(&ctx->os, &ctx->og);
			if (!ret)
            {
				break;
            }
			//int totalHeaderSize = (int)ctx->og.header_len + ctx->og.body_len;
			[self appendAccumulator:ctx data:(char *)ctx->og.header size:ctx->og.header_len];
            [self appendAccumulator:ctx data:(char *)ctx->og.body size:ctx->og.body_len];
		}
        
		ctx->isHeaderWritten = TRUE;
	}
    
	if (srcOffset < 0) {
		[self throwError:@"srcOffset < 0"];
	}
    
	if (srcLength > 0) {
		int numSamples = srcLength / 2;
        
		float **buffer = vorbis_analysis_buffer(&ctx->vd, numSamples);
        
		for (int i = 0; i < numSamples; i++) {
			int oi = (srcOffset + i) * 2;
			buffer[0][i] = ((pcmByteArray[oi+1]<<8) | (0x00ff&(int)pcmByteArray[oi]))/32768.f;
		}
        
		vorbis_analysis_wrote(&ctx->vd, numSamples);
	} else if (srcLength == 0) {
		return;
	} else  { // jSrcLength < 0
		// Tell Vorbis it's the end of stream
		vorbis_analysis_wrote(&ctx->vd, 0);
	}
    
	int eos = 0;
	while (vorbis_analysis_blockout(&ctx->vd, &ctx->vb) == 1) {
		//printf("vorbis_analysis_blockout\n");
		vorbis_analysis(&ctx->vb, NULL);
		vorbis_bitrate_addblock(&ctx->vb);
        
		while (vorbis_bitrate_flushpacket(&ctx->vd, &ctx->op)) {
			//printf("vorbis_bitrate_flushpacket\n");
            
			ogg_stream_packetin(&ctx->os, &ctx->op);
            
			while (!eos) {
				int result = ogg_stream_flush(&ctx->os, &ctx->og);
				if (!result)
					break;
                
				//int totalPageBytes = ctx->og.header_len + ctx->og.body_len;
				[self appendAccumulator:ctx data:(char *)ctx->og.header size:ctx->og.header_len];
				[self appendAccumulator:ctx data:(char *)ctx->og.body size:ctx->og.body_len];
                
				if (ogg_page_eos(&ctx->og)) {
					eos = 1;
				}
			}
		}
	}
}

// Release context
-(void)deleteEncoderContext:(int64_t)jEncoderContext {
	lastError = @"";
    
	OggVorbisEncoderContext *ctx = (OggVorbisEncoderContext *)jEncoderContext;
	if (ctx) {
		delete [] ctx->accumulator;
		ogg_stream_clear(&ctx->os);
		vorbis_block_clear(&ctx->vb);
		vorbis_dsp_clear(&ctx->vd);
		vorbis_comment_clear(&ctx->vc);
		vorbis_info_clear(&ctx->vi);
		delete ctx;
	}
}

+ (NSString *)convertAudioFile: (NSString*)audioFilePath
{
  OggEncoder *oggEncoder = [[OggEncoder alloc]init];
  NSTimeInterval timestamp = [NSDate timeIntervalSinceReferenceDate] * 1000.0;
  NSString *oggFilename = [NSString stringWithFormat: @"nvoqtranscriptionservice-%f.%@", timestamp, @"ogg"];
  NSString *oggFilepath = [NSTemporaryDirectory() stringByAppendingPathComponent:oggFilename];
  
  NSData *data = [NSData dataWithContentsOfFile:audioFilePath];
  
  int removeAppleWavHeaderLen = 44;
  NSData *subData = [data subdataWithRange:NSMakeRange(removeAppleWavHeaderLen, [data length] - removeAppleWavHeaderLen)];
  
  long dataLength = subData.length;
  NSMutableData *encodedOggData = [[NSMutableData alloc]init];
  long bufSize = 32768;
  for (long read = 0; read < dataLength; read+=bufSize) {
    // look for the last buffer, it needs to be smaller for NSMakeRange(position, numBytesToRead)
    if (read + bufSize >= dataLength) {
      bufSize = dataLength - read;
      [encodedOggData appendData: [oggEncoder encode:[subData subdataWithRange:NSMakeRange(read, bufSize)]]];
      break;
    }
    else {
      [encodedOggData appendData: [oggEncoder encode:[subData subdataWithRange:NSMakeRange(read, bufSize)]]];
    }
  }
  NSError *fwError = nil;
  [encodedOggData writeToFile:oggFilepath options:NSDataWritingAtomic error:&fwError];
  return oggFilepath;
}

@end

