/*
 * Copyright (c) 2000-2023 nVoq Incorporated
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

import UIKit

import AudioToolbox
import AVFoundation

class ViewController: UIViewController {
    var micAccessGranted = false
    //status flag
    var dictating = false
    
    //audio recording member variables
    static let kNumberRecordBuffers = 3
    var dataFormat: AudioStreamBasicDescription!
    var queue: AudioQueueRef?
    var buffers: AudioQueueBufferRef!
    var bufferByteSize: UInt32!
    var currentPacket: Int64!
    
    //this delegate listens to the recorder state and executes the
    //dictation appropriately.
    var delegate: AudioRecorderStateDelegate?
    
    @IBOutlet weak var button: UIButton!
    @IBOutlet weak var textArea: UITextView!
    
    override func viewDidLoad() {
        super.viewDidLoad()
        micAccessGranted = checkRecordPermission()
        // Do any additional setup after loading the view.
    }
    
    func checkRecordPermission() -> Bool {
        switch AVAudioSession.sharedInstance().recordPermission {
        case AVAudioSession.RecordPermission.granted:
            return true
        case AVAudioSession.RecordPermission.denied:
            print("No mic access granted")
            return false
        case AVAudioSession.RecordPermission.undetermined:
            AVAudioSession.sharedInstance().requestRecordPermission({ (granted) in
                self.micAccessGranted = granted
                
            })
            return false
        @unknown default:
            print("ERROR‼️:Unamaged record permission")
            return false
        }
    }
    
    //button click event handler
    @IBAction func buttonClicked(_ sender: Any) {
        if (self.dictating == false){
            self.dictating = true
            self.startDictation()
            button.setTitle("Stop", for: UIControl.State.normal)
        }else{
            self.dictating = false
            self.stopDictation()
            button.setTitle("Record", for: UIControl.State.normal)
        }
    }
    
    //Start listening for audio and start the dictation
    func startDictation(){
        print ("dictating...")
        startAudioListening()
    }
    //stop listening and end the dictation
    func stopDictation(){
        stop()
        print ("stopped dictating.")
    }
    
    
    func startAudioListening(){
        
        //when recording starts, create an instance of the delegate
        //that listens to audio recording events and performs the
        //dictation (start, send audio, receive text, stop) appropriately
        var dictationHelper = DictationHelper()
        dictationHelper.delegate = self
        delegate = dictationHelper
        
        //keep track of where we are in audio processing
        currentPacket = 0
        
        //create the audio description and initialize the values to zero
        var recordFormat = AudioStreamBasicDescription()
        
        //some old-school memory management - initialize record format
        //structure to zero
        memset(&recordFormat, 0, MemoryLayout<AudioStreamBasicDescription>.size)
        
        //set the parameters required to ask for 16kHz 16 bit audio
        recordFormat.mFormatID = kAudioFormatLinearPCM
        recordFormat.mChannelsPerFrame = 1
        recordFormat.mSampleRate = 16000
        recordFormat.mBitsPerChannel = 16
        recordFormat.mBytesPerFrame = UInt32(1*MemoryLayout<Int16>.stride)
        recordFormat.mFramesPerPacket = 1
        recordFormat.mBytesPerPacket = UInt32(1*MemoryLayout<Int16>.stride)
        recordFormat.mFormatFlags = kAudioFormatFlagIsSignedInteger | kAudioFormatFlagIsPacked
        var propSize = UInt32(MemoryLayout<AudioStreamBasicDescription>.size)
        
        // Get some audio format info
        AudioFormatGetProperty(kAudioFormatProperty_FormatInfo, 0, nil, &propSize, &recordFormat)
        
        // Create a new audio recording queue object, passing a reference to the callback
        // and a reference to this class instance
        AudioQueueNewInput(&recordFormat,
                           audioQueueInputCallback,
                           Unmanaged.passUnretained(self).toOpaque(), nil, nil, 0, &queue)
        
        // Allocate and enqueue buffers for 1/2 second audio chunks
        let bufferByteSize = UInt32(16000)
        //check to make sure we got a queue reference and unwrap it
        guard let unWrappedQueue = queue else {
            return
        }
        for _ in 0...ViewController.kNumberRecordBuffers {
            var buffer: AudioQueueBufferRef?
            AudioQueueAllocateBuffer(unWrappedQueue, bufferByteSize, &buffer)
            //check to make sure we have a non-nil buffer
            guard let unWrappedBuffer = buffer else {
                return
            }
            AudioQueueEnqueueBuffer(unWrappedQueue, unWrappedBuffer, 0, nil)
        }
        
        // Start the queue
        AudioQueueStart(unWrappedQueue, nil)
        delegate?.onAudioRecordingStart() //let our delegate know we succeeded
        return
    }
    
    func stop() {
        dictating = false
    }
    
    func interrupt() {
        guard let queue = queue else {
            print("Queue is nil")
            return
        }
        dictating = false
        AudioQueueStop(queue, false)
        NotificationCenter.default.removeObserver(self)
        return
    }
}

func audioQueueInputCallback(inUserData: UnsafeMutableRawPointer?,
                             inQueue: AudioQueueRef,
                             inBuffer: AudioQueueBufferRef,
                             inStartTime: UnsafePointer<AudioTimeStamp>,
                             inNumPackets: UInt32,
                             inPacketDesc: UnsafePointer<AudioStreamPacketDescription>?) {
    
    guard let inUserData = inUserData else {
        return
    }
    
    //recorder is a reference back to the instance of view controller
    let recorder = Unmanaged<ViewController>.fromOpaque(inUserData).takeUnretainedValue()
    guard let queue = recorder.queue else {
        print("InputCallBack - audio queue is nil")
        return
    }
    //var inNumPacketsVar = inNumPackets
    
    //did we get any data?  if so, process it...
    if recorder.dictating && inNumPackets > 0 {
        
        // Convert raw data to NSData
        let audioData = NSData(bytesNoCopy: inBuffer.pointee.mAudioData, length: Int(inBuffer.pointee.mAudioDataByteSize), freeWhenDone: false)
        
        // Send the audio bytes to the delegate
        recorder.delegate?.onAudioRecordingData(audioData: audioData)
        recorder.currentPacket = recorder.currentPacket + Int64(inNumPackets)
    }
    
    //If we are still recording audio & dictating
    //put the buffer back in the queue
    if recorder.dictating {
        AudioQueueEnqueueBuffer(queue, inBuffer, 0, nil)
    } else if !recorder.dictating { //otherwise, shut things down
        recorder.dictating = false
        recorder.delegate?.onAudioRecordingStop()
        if AudioQueueStop(queue, true) != errSecSuccess {
            print("audioQueueInputCallback - Queue did not stop")
        }
        let resultDispose = AudioQueueDispose(queue, true)
        if resultDispose != errSecSuccess {
            AudioQueueDispose(queue, true)
        }
        recorder.delegate?.onAudioQueueDisposed()
    }
}

extension ViewController: DictationHelperDelegate {
    func onTextReceived(text: String) {
        DispatchQueue.main.async {
            self.textArea.text = text
        }
    }
}
